//gcc qc01.c -Wall -g -o qc01 -L/home/sbester/server/5.0/mysql-5.0.46-linux-i686/lib -I/home/sbester/server/5.0/mysql-5.0.46-linux-i686/include   -lmysqlclient_r -lz -lpthread



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>
#include "mysql.h"


#define TESTTIME (3600)
#define NUMTHREADS (5)
#define NUMTABLES 2
#define ROWSPERTABLE 5000000
#define QUERYCACHESIZE 1024*1024*1024
#define INSERTTHREADSLEEP 300
#define PERCENTTABLESMODIFY 50

char host[]="127.0.0.1";
char username[]="root";
char password[]="";
char database[]="test";
int port=33046;
pthread_t pthreads[NUMTHREADS];
pthread_t inserter;

int threaddone=0;
unsigned long long total_selects=0;

void *selecting_thread(void *arg);
void *inserting_thread(void *arg);

MYSQL *db_connect(MYSQL *dbc);
int db_query(MYSQL *dbc,char *sql,int showresults);


int main(int argc, const char *argv[])
{
	MYSQL *dbc=NULL;
	MYSQL_RES *r=NULL;
	MYSQL_ROW w;
	int res=0;
	unsigned int i=0,err=0;

	time_t timestart=0,timenow=0;
        srand48(time(NULL));
	
	char shortquery[1024]={0};
	char *c=NULL;
	my_init();
	if (!(dbc = mysql_init(NULL)))
	{
		printf("mysql_init\n");
		dbc=NULL;
		goto threadexit;
	}
	else
	{
		if (!mysql_real_connect(dbc,host,username,password,database,port, NULL, CLIENT_FOUND_ROWS|CLIENT_MULTI_STATEMENTS|CLIENT_MULTI_RESULTS))
		{
			printf("mysql_real_connect failed: %s (%d)", mysql_error(dbc),mysql_errno(dbc));
			dbc=NULL;
			goto threadexit;
		}
	}

	printf("running initializations..\n");
	
	c=shortquery;
	c+=sprintf(c,"%s","flush status");
	db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"%s","flush tables");
	db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"%s","reset query cache");
	db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"set global query_cache_size=%d",QUERYCACHESIZE);
	//db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"%s","set global query_cache_type=1");
	db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"%s","set global query_cache_min_res_unit=4096");
	db_query(dbc,shortquery,1);

	c=shortquery;
	c+=sprintf(c,"%s","set global max_allowed_packet=1024*1024*1024");
	db_query(dbc,shortquery,1);

        c=shortquery;
        c+=sprintf(c,"%s","drop database if exists `qc01`");
        db_query(dbc,shortquery,1);
        
        c=shortquery;
        c+=sprintf(c,"%s","create database `qc01`");
        db_query(dbc,shortquery,1);
        
        
        printf("creating %d tables\n",NUMTABLES);
        for(i=0;i<NUMTABLES;i++)
        {
                c=shortquery;
                c+=sprintf(c,"drop table if exists `q%04d`",i);
                db_query(dbc,shortquery,1);
                
                c=shortquery;
                c+=sprintf(c,"create table `qc01`.`q%04d`(`a` mediumint unsigned not null primary key, b int default 0)engine=myisam",i);
                db_query(dbc,shortquery,1);
        }
        
        printf("populating first table with %d random values\n",ROWSPERTABLE);
        for(i=0;i<ROWSPERTABLE;i++)
        {
                c=shortquery;
                c+=sprintf(c,"insert ignore into `qc01`.`q0000`(`a`) values (%lu)",lrand48() % (ROWSPERTABLE*2));
                db_query(dbc,shortquery,1);
        }

        printf("copying first table to other tables\n");
        
        for(i=1;i<NUMTABLES;i++)
        {
                c=shortquery;
                c+=sprintf(c,"insert into `qc01`.`q%04d` select * from `qc01`.`q0000`",i);
                db_query(dbc,shortquery,1);
                printf("done `qc01`.`q%04d`\n",i);
        }
        sleep(1);

        c=shortquery;
	c+=sprintf(c,"%s","flush tables");
	db_query(dbc,shortquery,1);
	//mysql_close(dbc);

	printf("about to spawn %d select threads\n",NUMTHREADS);
	for (i=0;i<NUMTHREADS;i++)
	{
		err=pthread_create(&pthreads[i], NULL, selecting_thread, (void *)i);
		if(err!=0)
		{
			printf("error spawning thread %lu, pthread_create returned %lu\n",(unsigned long)i,(unsigned long)err);
		}
	        printf(".");
	}
	printf("\n");
	printf("completed spawning new threads\n");
        err=pthread_create(&inserter,NULL,inserting_thread,0);
        if(err!=0)
                printf("error spawning inserting thread.  pthread_create returned %lu\n",(unsigned long)err);
        

	printf("testcase is now running\n");

	timestart=time(NULL);
	timenow=time(NULL);
	struct timeval tim;
        
	printf("elapsed time,selects since last printout,Qcache_free_blocks,Qcache_free_memory,Qcache_hits,Qcache_inserts,Qcache_lowmem_prunes,Qcache_not_cached,Qcache_queries_in_cache,Qcache_total_blocks\n");
	gettimeofday(&tim, NULL);
	double t1=tim.tv_sec+(tim.tv_usec/1000000.0);
	double t2=0;
	unsigned long long old_selects=0;
	for(i=0;(timenow-timestart) < TESTTIME;timenow=time(NULL))
	{
		sleep(3);
		gettimeofday(&tim, NULL);
                t2=tim.tv_sec+(tim.tv_usec/1000000.0);
                printf("%.6lf,%lu,",(float)(t2-t1),(unsigned long)(total_selects-old_selects));
                old_selects=total_selects;
                
                res = mysql_query(dbc,"show global status like '%qc%'");
	        if(res != 0)
	        {
		        printf("\ncan't get qcache status: %d (%s)\n",mysql_errno(dbc),mysql_error(dbc));
		        exit(1);
	        }
		r = mysql_store_result(dbc);
		/*
		print this on a single line.
		Qcache_free_blocks      13668
                Qcache_free_memory      43719888
                Qcache_hits     1885
                Qcache_inserts  297753
                Qcache_lowmem_prunes    80296
                Qcache_not_cached       7
                Qcache_queries_in_cache 162046
                Qcache_total_blocks     337860

		*/
		while ((w = mysql_fetch_row(r)))
                {
                        printf("%12s,",w[1]);
                }
                printf("\n");
                mysql_free_result(r);
                fflush(stdout);
	}
	threaddone=1;

	printf("waiting for worker threads to finish...\n");

	for (i=0;i<NUMTHREADS;i++)
	{
		pthread_join(pthreads[i], NULL);
	}
	pthread_join(inserter,NULL);

	exit(0);
threadexit:
	exit(-1);
}


void *selecting_thread(void *arg)
{
	MYSQL *dbc=NULL;
	int cancelstate=0;
	char shortquery[1024]={0};
	char *c=NULL;
	
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,&cancelstate);

	if (!(dbc = mysql_init(NULL)))
	{
		printf("mysql_init\n");
		dbc=NULL;
		goto threadexit;
	}
	else
	{
		if (!mysql_real_connect(dbc,host,username,password,database,port, NULL, CLIENT_FOUND_ROWS|CLIENT_MULTI_STATEMENTS|CLIENT_MULTI_RESULTS))
		{
			printf("mysql_real_connect failed: %s (%d)", mysql_error(dbc),mysql_errno(dbc));
			dbc=NULL;
		}
	}

	
	while(!threaddone && dbc!=NULL)
	{
                c=shortquery;
		c+=sprintf(c,"select * from `qc01`.`q%04d` where `a`=%lu",(int)(lrand48()%NUMTABLES),lrand48()%(ROWSPERTABLE*2));
		db_query(dbc,shortquery,1);
		total_selects++;
	}
threadexit:
	if(dbc)
	        mysql_close(dbc);
	mysql_thread_end();
	pthread_exit(0);
}



void *inserting_thread(void *arg)
{
	MYSQL *dbc=NULL;
	int cancelstate=0;
	char shortquery[1024]={0};
	char *c=NULL;
	unsigned int i=0;
	
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,&cancelstate);

	if (!(dbc = mysql_init(NULL)))
	{
		printf("mysql_init\n");
		dbc=NULL;
		goto threadexit;
	}
	else
	{
		if (!mysql_real_connect(dbc,host,username,password,database,port, NULL, CLIENT_FOUND_ROWS|CLIENT_MULTI_STATEMENTS|CLIENT_MULTI_RESULTS))
		{
			printf("mysql_real_connect failed: %s (%d)", mysql_error(dbc),mysql_errno(dbc));
			dbc=NULL;
		}
	}


	while(!threaddone && dbc!=NULL)
	{
                sleep(INSERTTHREADSLEEP);
                //printf("about to make %d random inserts to about 50%% of the tables\n",NUMTABLES/10);
                //c=shortquery;
                //c+=sprintf(c,"%s","show global status like '%qc%'");
                //db_query(dbc,shortquery,2);
                
                for(i=0;i<NUMTABLES/(100/PERCENTTABLESMODIFY);i++)
                {
                        c=shortquery;
                        c+=sprintf(c,"insert ignore into `qc01`.`q%04d`(`a`) values (%lu)",(int)(lrand48()%NUMTABLES),(lrand48() % (ROWSPERTABLE*2)));
			//c+=sprintf(c, "update `qc01`.`q%04d` set b=%lu where a=%lu",(int)(lrand48()%NUMTABLES),(lrand48() % (ROWSPERTABLE*2)),(lrand48() % (ROWSPERTABLE*2)));
                        db_query(dbc,shortquery,1);
		}
		
		//c=shortquery;
                //c+=sprintf(c,"%s","show global status like '%qc%'");
                //db_query(dbc,shortquery,2);
	}
threadexit:
	if(dbc)
	        mysql_close(dbc);
	mysql_thread_end();
	pthread_exit(0);
}



int db_query(MYSQL *dbc,char *sql,int showresults)
{
	int res=0;
	MYSQL_RES *r=NULL;
	MYSQL_ROW w;
	MYSQL_FIELD *field=NULL;
	int i=0,moreresult=0;
	if(NULL == dbc) return 0;
	res = mysql_query(dbc,sql);
	if(res != 0 && showresults > 0)
	{
		printf("query failed '%s' : %d (%s)\n",sql,mysql_errno(dbc),mysql_error(dbc));
		return 0;
	}

	do
	{
		r = mysql_store_result(dbc);
		if(r)
		{
			unsigned int numfields = mysql_num_fields(r);
			//unsigned int numrows=mysql_num_rows(r);
			while((field = mysql_fetch_field(r)))
			{
					//print metadata information about each field
					if(showresults > 1)
					{
						printf("%s	",field->name);
					}
			}
			if(showresults > 1)
			{
				printf("\n------------------------------------\n");
			}

			while ((w = mysql_fetch_row(r)))
			{
				for(i = 0; i < numfields; i++)
				{
					//print each field here
					if(showresults > 1)
					{
						printf("%s\t",w[i]);
					}
				}

				if(showresults > 1)
				{
						printf("\n");
				}
			}
			if(showresults > 1)
			{
				printf("\n");
			}
			mysql_free_result(r);
		}
		else //no rows returned. was it a select?
		{
			if(mysql_field_count(dbc) > 0 && showresults > 0)
			{
				printf("No results for '%s'.  (%d) - %s\n",sql,mysql_errno(dbc),mysql_error(dbc));
				return 0;
			}
			else //it could have been some insert/update/delete
			{
				//this is successful query
			}
		}
		moreresult=mysql_next_result(dbc);
		if(moreresult > 0 && showresults > 0)
		{
			printf("mysql_next_result returned %d, mysql error %s, (%d)\n",moreresult,mysql_error(dbc),mysql_errno(dbc));
			break;
		}
	} while (0==moreresult);
	return 1;
}

